/*-------------------------------------------------------------------------
 * Copyright (c) 1999-2003 Kenneth W. Sodemann (stuffle@mac.com)
 *-------------------------------------------------------------------------
 * probupdate_dlg
 *
 * Synopsis:
 *   Handle updates to, and viewing of existing problem reports.
 *
 * $Id: probupdate_dlg.c,v 1.62 2003/12/20 14:50:01 stuffle Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to
 * Free Software Foundation, Inc.
 * 59 Temple Place, Suite 330
 * Boston, MA  02111-1307  USA
 *
 *-------------------------------------------------------------------------
 */
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <sys/stat.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <gnome.h>
#include <gconf/gconf-client.h>
#include <libpq-fe.h>

#include "audit_trail.h"
#include "db_utils.h"
#include "defs.h"
#include "error_chks.h"
#include "gstr_utils.h"
#include "gtkutils.h"
#include "mainwin.h"
#include "probupdate_dlg.h"
#include "util.h"

/*
 * Define the names of some of the key widgets.
 */
#define PROJECT_LABEL          "project_label"
#define VERSION_LABEL          "version_label"
#define PRBTYPE_LABEL          "prbtype_label"
#define SUBMTTR_LABEL          "submttr_label"

#define RESPONSIBLE_MENU       "responsible_menu"
#define SUBJECT_ENTRY          "subject_entry"
#define STATUS_MENU            "status_menu"
#define SEVERITY_MENU          "severity_menu"
#define FIX_VER_MENU           "fix_ver_menu"

#define PK_LIST                "pk_list"

#define PROJECT_ENTRY          "project_entry"
#define SUBMITTER_ENTRY        "submitter_entry"
#define RESPONSIBLE_COMBO      "responsible_combo"
#define STATUS_LIST_VIEW       "status_list_view"
#define PROBLEM_TYPE_ENTRY     "problem_type_entry"
#define SEVERITY_LIST_VIEW     "severity_list_view"
#define DESCR_EDIT             "descr_edit"
#define FIX_DESCR_EDIT         "fix_descr_edit"
#define AUDIT_TRAIL_EDIT       "aud_tr_edit"
#define RESPONSIBLE_LIST       "responsible_list"
#define OPEN_VER_ENTRY         "open_ver_entry"
#define FIX_VER_COMBO          "fix_ver_combo"
#define FIX_VER_ENTRY          "fix_ver_entry"
#define CHANGE_LOG_TEXT        "change_log_text"
#define FIX_TAB_LBL            "fix_tab_label"
#define CHGLOG_TAB_LBL         "chglog_tab_label"
#define FIX_TAB_CHAR_CNT       "fix_tab_char_cnt"
#define CHGLOG_TAB_CHAR_CNT    "chglog_tab_char_cnt"

#define FIX_DESCR              _("Fix Description")
#define FIX_DESCR_STAR         _("Fix Description*")
#define CHGLOG_DESCR           _("ChangeLog Text")
#define CHGLOG_DESCR_STAR      _("ChangeLog Text*")

#define RESP_NOBODY_NAME       _("<None>")
#define RESP_NOBODY_ID         "" 

#define CLIST_START   7
#define CLIST_END     10 

/*
 * This "locking" is voluntary, but it seems to work much nicer than the
 * current levels of locking available from PostgreSQL.  The basic idea
 * here is to run the CREATE_LOCK query, then check and make sure that 
 * the locker is you.  If the locker is someone else, they had it first.
 *
 * REMOVE_LOCK must be run before the dialog exits.
 */
#define CREATE_LOCK "\
UPDATE problem_report \
   SET locker_id = getpgusername () \
 WHERE locker_id is NULL \
   AND problem_num = %d "
#define SELECT_LOCKER "\
SELECT pr.locker_id, \
       p.last_name || ', ' || p.first_name || ' (' || pr.locker_id || ')' \
  FROM problem_report pr, person p \
 WHERE p.login_id = pr.locker_id \
   AND problem_num = %d "
#define REMOVE_LOCK "\
UPDATE problem_report \
   SET locker_id = NULL \
 WHERE locker_id = getpgusername () \
   AND problem_num = %d "

/*
 * Get all of the information that we need for the problem report.
 */
#define PRJ_NUM_POS          0
#define PRJ_NAME_POS         1 
#define SUBMITTER_ID_POS     2 
#define RESPONSIBLE_ID_POS   3
#define TITLE_POS            4
#define DESCR_POS            5
#define FIX_POS              6
#define SEVERITY_NUM_POS     7
#define STATUS_NUM_POS       8
#define PROB_TYPE_POS        9
#define OPEN_VER_POS         10
#define FIX_VER_POS          11
#define CHANGE_LOG_TEXT_POS  12
#define SELECT_PR_DATA "\
SELECT pr.project_num, prj.name, \
       p.last_name||', '|| p.first_name||' ('||pr.submitter_id||')', \
       pr.responsible_id, pr.title, pr.descr, pr.fix, pr.severity_num, \
       pr.status_num, pt.name, pr.open_version, pr.close_version, \
       pr.change_log_text \
  FROM problem_report pr, project prj, person p, problem_type pt \
 WHERE pt.type_num = pr.type_num \
   AND p.login_id = pr.submitter_id \
   AND prj.project_num = pr.project_num \
   AND pr.problem_num = %d "

#define FULL_NAME_POS        4
#define SELECT_RESPONSIBLE "\
SELECT 0, 1, 2, r.login_id, \
       p.last_name||', '||p.first_name||' ('||r.login_id||')' \
  FROM responsible r, person p, problem_report pr \
 WHERE p.login_id = r.login_id \
   AND r.project_num = pr.project_num \
   AND pr.problem_num = %d \
 ORDER BY p.last_name, p.first_name, r.login_id  "

#define NAME_POS           0
#define PK_NUM_POS         1
#define ORDER_POS          2
#define SELECT_STATUSES "\
SELECT s.name, s.status_num, s.order_num \
  FROM status s, problem_report pr \
 WHERE (s.status_num = pr.status_num \
        OR s.active = TRUE) \
   AND pr.problem_num = %d \
 ORDER BY s.order_num "

#define SELECT_SEVERITIES "\
SELECT s.name, s.severity_num, s.order_num \
  FROM severity s, problem_report pr \
 WHERE (s.severity_num = pr.severity_num \
        OR s.active = TRUE) \
   AND pr.problem_num = %d \
 ORDER BY s.order_num "

#define VERSION_NUM_POS      1 
#define VERSION_TXT_POS      2 
#define SELECT_VERSIONS "\
SELECT pv.order_num, pv.version_num, pv.version_text \
  FROM project_version pv, problem_report pr \
 WHERE pv.active = TRUE \
   AND pv.project_num = pr.project_num \
   AND pr.problem_num = %d \
 UNION \
SELECT pv.order_num, pv.version_num, pv.version_text \
  FROM project_version pv, problem_report pr \
 WHERE pv. project_num = pr.project_num \
   AND pv.version_num = pr.close_version \
   AND pr.close_version is not NULL \
   AND pr.problem_num = %d \
 ORDER BY 1  "

#define SELECT_VER_TEXT "\
SELECT version_text, version_num \
  FROM project_version \
 WHERE version_num = %d \
   AND project_num = %d "

/*
 * The VERSION_NAME_PK_LIST stores the name/pk combinations that are
 * used in the FIX_VER_COMBO.
 */
#define VERSION_NAME_PK_LIST    "version_name_pk_list"

/*
 * 'Null' version display data (for the version related stuff).
 */
#define NULL_NAME           "<None>"
#define NULL_NAME_SIZE      6

enum {
   ATTRIB_NAME = 0,
   ATTRIB_PK,
   NUM_ATTRIB_COLS
};

/*----------------------------------------------------------------------
 *  Local Function Prototypes.
 *--------------------------------------------------------------------*/
static void     check_chglog_tab_text (GtkTextBuffer *buff,
                                       gpointer       user_data);
static void     check_fix_tab_text (GtkTextBuffer *buff,
                                    gpointer       user_data);
static void     draw_audit_trail (GtkTextView *view, 
                                 PGconn      *conn,
                                 gint         pr_pk);
static void     on_pr_update_dlg_show (GtkWidget *widget,
                                      gpointer   user_data);
static void     on_pr_update_dlg_destroy (GObject  *object,
                                          gpointer  user_data);
static void     on_cancel_button_clicked (GtkDialog *dlg);
static void     on_pr_update_dlg_clicked (GtkDialog *dlg,
                                          gint       btn_cd,
                                          gpointer   user_data);
static gboolean submitter_access_only (PGconn  *conn, 
                                       gint     prnum);


static void 
draw_audit_trail (GtkTextView *view, 
                  PGconn      *conn,
                  gint         pr_pk)
{
   GtkTextBuffer *buff;
   GtkTextIter    end;
   GString       *sql_buffer;
   PGresult      *res;
   gint           i, n;

   buff = gtk_text_view_get_buffer (view);
   sql_buffer = g_string_new ("");
   g_string_printf (sql_buffer, PR_SUBMIT_DATE_QUERY, pr_pk);
   res = PQexec (conn, sql_buffer->str);
   if (chk_sql_error (res, "Getting open line"))
      {
      g_assert (PQntuples (res) == 1);
      gtk_text_buffer_set_text (buff, PQgetvalue (res, 0, 0), -1);
      gtk_text_buffer_get_end_iter (buff, &end);
      gtk_text_buffer_insert (buff, &end, "\n", -1);
      }
   PQclear (res);

   g_string_printf (sql_buffer, SELECT_AUDIT_TRAIL, pr_pk, pr_pk);
   res = PQexec (conn, sql_buffer->str);
   if (chk_sql_error (res, "Getting Audit Trail Data"))
      {
      n = PQntuples (res);
      for (i = 0; i < n; i++)
         {
         gtk_text_buffer_insert (buff, &end, PQgetvalue (res, i, 0), -1);
         gtk_text_buffer_insert (buff, &end, "\n", -1);
         }
      }

   g_string_free (sql_buffer, TRUE);
   PQclear (res);
}


static gboolean
submitter_access_only (PGconn *conn,
                       gint    prnum)
{
   PGresult   *res;
   GString    *sql;
   gint        n = 0;

   sql = g_string_new ("select count (*) ");
   sql = g_string_append (sql, "from submitter s, problem_report pr, ");
   sql = g_string_append (sql, "project p ");
   sql = g_string_append (sql, "where p.project_num = pr.project_num ");
   sql = g_string_append (sql, "and p.login_id <> getpgusername() ");
   sql = g_string_append (sql, "and s.project_num = pr.project_num ");
   sql = g_string_append (sql, "and s.login_id = getpgusername() ");
   g_string_append_printf (sql, "and pr.problem_num = %d ", prnum);
   sql = g_string_append (sql, "and not exists ( ");
   sql = g_string_append (sql, "select 'x' from responsible ");
   sql = g_string_append (sql, "where login_id = s.login_id ");
   sql = g_string_append (sql, "and project_num = s.project_num) ");

   res = PQexec (conn, sql->str);
   if (chk_sql_error (res, _("Determine if submitter only")))
      {
      n = atoi (PQgetvalue (res, 0, 0));
      }

   PQclear (res);
   g_string_free (sql, TRUE);

   return (n > 0);
}


static void 
on_pr_update_dlg_show (GtkWidget *widget, gpointer  user_data)
{
   GtkLabel         *project_lbl;
   GtkLabel         *open_ver_lbl;
   GtkLabel         *pr_type_lbl;
   GtkLabel         *submitter_lbl;
   GtkOptionMenu    *responsible_menu;
   GtkOptionMenu    *severity_menu;
   GtkOptionMenu    *status_menu;
   GtkOptionMenu    *fix_ver_menu;
   GtkEntry         *subject_entry;
   GtkTextView      *descr_text;
   GtkTextView      *fix_text;
   GtkTextView      *change_log_text;
   GtkTextView      *au_tr_text;
   GtkTextBuffer    *buff;
   GtkLabel         *lbl;
   GString          *sql_buffer;
   GString          *id;
   GString          *resp_id;
   PGresult         *main_res = NULL;  /* main PR select result */
   PGresult         *tmp_res;          /* temp results for subselects */
   GList            *pk_list = NULL;
   gboolean          in_error = FALSE;
   gint              i, n;
   gint              sel;
   gint              txt_len;
   PGconn           *conn;
   gint              pk_num;  /* the pk_num of the problem report */
   gint              pk;      /* temp PK used for populating menus */
   gint              targ_pk; /* target PK used to init menu values */
   gint              prj_pk;  /* the project */
   GtkWidget        *msg_dlg;
   GString          *msg_str;
   GtkWidget        *menu;
   GtkWidget        *item;

   /*
    * Get pointers to the widgets that we need to fill.
    */
   project_lbl = GTK_LABEL (lookup_widget (widget, PROJECT_LABEL));
   open_ver_lbl = GTK_LABEL (lookup_widget (widget, VERSION_LABEL));
   pr_type_lbl = GTK_LABEL (lookup_widget (widget, PRBTYPE_LABEL));
   submitter_lbl = GTK_LABEL (lookup_widget (widget, SUBMTTR_LABEL));

   responsible_menu = GTK_OPTION_MENU (lookup_widget (widget,
                                                      RESPONSIBLE_MENU));
   severity_menu = GTK_OPTION_MENU (lookup_widget (widget,
                                                   SEVERITY_MENU));
   status_menu = GTK_OPTION_MENU (lookup_widget (widget, STATUS_MENU));
   fix_ver_menu = GTK_OPTION_MENU (lookup_widget (widget, FIX_VER_MENU));
   subject_entry = GTK_ENTRY (lookup_widget (widget, SUBJECT_ENTRY));

   descr_text = GTK_TEXT_VIEW (lookup_widget (widget, DESCR_EDIT));
   fix_text = GTK_TEXT_VIEW (lookup_widget (widget, FIX_DESCR_EDIT));
   change_log_text = GTK_TEXT_VIEW (lookup_widget (widget, CHANGE_LOG_TEXT));
   au_tr_text = GTK_TEXT_VIEW (lookup_widget (widget, AUDIT_TRAIL_EDIT));

   g_assert (project_lbl != NULL);
   g_assert (open_ver_lbl != NULL);
   g_assert (pr_type_lbl != NULL);
   g_assert (submitter_lbl != NULL);
   
   g_assert (responsible_menu != NULL);
   g_assert (severity_menu != NULL);
   g_assert (status_menu != NULL);
   g_assert (subject_entry != NULL);
   g_assert (fix_ver_menu != NULL);

   g_assert (descr_text != NULL);
   g_assert (fix_text != NULL);
   g_assert (change_log_text != NULL);
   g_assert (au_tr_text != NULL);


   conn = g_object_get_data (G_OBJECT (widget), MY_DB_CONN);
   g_assert (conn != NULL);
   pk_num = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (widget), PK_NUM));
   sql_buffer = g_string_new ("");

   /*
    * Check the locking.
    */
   START_TRANSACTION (conn);
   g_string_printf (sql_buffer, CREATE_LOCK, pk_num);
   tmp_res = PQexec (conn, sql_buffer->str);
   in_error = !chk_sql_error (tmp_res, "Creating voluntary lock");
   if (in_error)
      {
      ROLLBACK (conn);
      }
   else
      {
      COMMIT (conn);
      }
   PQclear (tmp_res);

   if (!in_error)
      {
      g_string_printf (sql_buffer, SELECT_LOCKER, pk_num);
      tmp_res = PQexec (conn, sql_buffer->str);
      in_error = !chk_sql_error (tmp_res, "Checking the lock status");
      if (!in_error)
         {
         if (strcmp (PQuser (conn), PQgetvalue (tmp_res, 0, 0)))
            {
            msg_str = g_string_new ("");
            g_string_printf
               (msg_str,
                _("The selected problem report is locked by: %s\n"),
                PQgetvalue (tmp_res, 0, 1));
            msg_str = g_string_append
               (msg_str,
                _("The problem report will be opened for viewing only."));
            msg_dlg = gtk_message_dialog_new (GTK_WINDOW (widget), 0,
                                              GTK_MESSAGE_INFO,
                                              GTK_BUTTONS_OK,
                                              msg_str->str);
            gtk_dialog_run (GTK_DIALOG (msg_dlg));
            gtk_dialog_set_response_sensitive (GTK_DIALOG (widget),
                                               GTK_RESPONSE_OK,
                                               FALSE);
            gtk_dialog_set_response_sensitive (GTK_DIALOG (widget),
                                               GTK_RESPONSE_APPLY,
                                               FALSE);
            gtk_dialog_set_default_response (GTK_DIALOG (widget),
                                             GTK_RESPONSE_CANCEL);
            gtk_widget_set_sensitive (GTK_WIDGET (status_menu), FALSE);
            gtk_widget_set_sensitive (GTK_WIDGET (severity_menu), FALSE);
            gtk_widget_set_sensitive (GTK_WIDGET (fix_ver_menu), FALSE);
            gtk_widget_set_sensitive (GTK_WIDGET (responsible_menu), FALSE);
            gtk_widget_set_sensitive (GTK_WIDGET (descr_text), FALSE);
            gtk_widget_set_sensitive (GTK_WIDGET (fix_text), FALSE);
            gtk_widget_set_sensitive (GTK_WIDGET (change_log_text), FALSE);
            gtk_widget_set_sensitive (GTK_WIDGET (subject_entry), FALSE);
            }
         }
      PQclear (tmp_res);
      }

   /*
    * Check for submitter only access.
    */
   if (submitter_access_only (conn, pk_num))
      {
      gtk_widget_set_sensitive (GTK_WIDGET (status_menu), FALSE);
      gtk_widget_set_sensitive (GTK_WIDGET (severity_menu), FALSE);
      gtk_widget_set_sensitive (GTK_WIDGET (fix_ver_menu), FALSE);
      gtk_widget_set_sensitive (GTK_WIDGET (responsible_menu), FALSE);
      gtk_widget_set_sensitive (GTK_WIDGET (fix_text), FALSE);
      gtk_widget_set_sensitive (GTK_WIDGET (change_log_text), FALSE);
      }
      
   /*
    * Start by querying back all of the data for the current problem report.
    * The results of this query will be stored in main_res for the duration
    * of the function call.
    *
    * Populate the widgets that can be populated directly.
    *
    * The menu widgets need to be populated with all of their possible 
    * values before we can set their value.  Thus, we set the value of
    * those widgets as we initialize their possible values.
    */
   START_TRANSACTION (conn);
   if (!in_error)
      {
      g_string_printf (sql_buffer, SELECT_PR_DATA, pk_num);
      main_res = PQexec (conn, sql_buffer->str);
      if (chk_sql_error (main_res, "Populating PR Edit Dialog"))
         {
         g_assert (PQntuples (main_res) == 1);
         prj_pk = atoi (PQgetvalue (main_res, 0, PRJ_NUM_POS));
         gtk_label_set_text (project_lbl, 
                             PQgetvalue (main_res, 0, PRJ_NAME_POS)); 
         gtk_label_set_text (pr_type_lbl,
                             PQgetvalue (main_res, 0, PROB_TYPE_POS));
         gtk_label_set_text (submitter_lbl, 
                             PQgetvalue (main_res, 0, SUBMITTER_ID_POS)); 
         if (PQgetisnull (main_res, 0, OPEN_VER_POS))
            {
            gtk_label_set_text (open_ver_lbl, NULL_NAME);
            }
         else
            {
            pk = atoi (PQgetvalue (main_res, 0, OPEN_VER_POS));
            g_string_printf (sql_buffer, SELECT_VER_TEXT, pk, prj_pk);
            tmp_res = PQexec (conn, sql_buffer->str);
            g_assert (PQntuples (tmp_res) == 1);
            if (chk_sql_error (tmp_res, "Get open version text"))
               {
               gtk_label_set_text (open_ver_lbl, 
                                   PQgetvalue (tmp_res, 0, NAME_POS));
               }
            PQclear (tmp_res);
            }

         gtk_entry_set_text (subject_entry, 
                             PQgetvalue (main_res, 0, TITLE_POS));
         
         buff = gtk_text_view_get_buffer (descr_text);
         gtk_text_buffer_set_text (buff,
                                   PQgetvalue (main_res, 0, DESCR_POS), -1);
         gtk_text_buffer_set_modified (buff, FALSE);
         
         buff = gtk_text_view_get_buffer (fix_text);
         gtk_text_buffer_set_text (buff,
                                   PQgetvalue (main_res, 0, FIX_POS), -1);
         gtk_text_buffer_set_modified (buff, FALSE);
         txt_len = gtk_text_buffer_get_char_count (buff);
         lbl = GTK_LABEL (lookup_widget (widget, FIX_TAB_LBL));
         g_assert (lbl != NULL);
         if (txt_len == 0)
            {
            gtk_label_set_text (lbl, FIX_DESCR);
            }
         else
            {
            gtk_label_set_text (lbl, FIX_DESCR_STAR);
            }
         g_object_set_data (G_OBJECT (widget), FIX_TAB_CHAR_CNT,
                            GINT_TO_POINTER (txt_len));
         
         buff = gtk_text_view_get_buffer (change_log_text);
         gtk_text_buffer_set_text (buff,
               PQgetvalue (main_res, 0, CHANGE_LOG_TEXT_POS), -1);
         gtk_text_buffer_set_modified (buff, FALSE);
         txt_len = gtk_text_buffer_get_char_count (buff);
         lbl = GTK_LABEL (lookup_widget (widget, CHGLOG_TAB_LBL));
         g_assert (lbl != NULL);
         if (txt_len == 0)
            {
            gtk_label_set_text (lbl, CHGLOG_DESCR);
            }
         else
            {
            gtk_label_set_text (lbl, CHGLOG_DESCR_STAR);
            }
         g_object_set_data (G_OBJECT (widget), CHGLOG_TAB_CHAR_CNT,
                            GINT_TO_POINTER (txt_len));
         
         draw_audit_trail (au_tr_text, conn, pk_num);
         }
      else
         {
         in_error = TRUE;
         }
      }

   /*
    * Fill in the status menu.
    */
   if (!in_error)
      {
      g_string_printf (sql_buffer, SELECT_STATUSES, pk_num);
      tmp_res = PQexec (conn, sql_buffer->str);
      sel = 0;
      targ_pk = atoi (PQgetvalue (main_res, 0, STATUS_NUM_POS));
      if (chk_sql_error (tmp_res, "Filling the statuses list"))
         {
         menu = gtk_menu_new ();
         n = PQntuples (tmp_res);
         for (i = 0; i < n; i++)
            {
            pk = atoi (PQgetvalue (tmp_res, i, PK_NUM_POS));
            sel = (pk == targ_pk) ? i : sel;
            pk_list = g_list_append (pk_list, GINT_TO_POINTER (pk));
            item = gtk_menu_item_new_with_label
               (PQgetvalue (tmp_res, i, NAME_POS));
            gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
            gtk_widget_show (item);
            }
         g_object_set_data (G_OBJECT (status_menu), PK_LIST, pk_list);
         pk_list = NULL;
         gtk_option_menu_set_menu (status_menu, menu);
         gtk_option_menu_set_history (status_menu, sel);
         }
      else
         {
         in_error = TRUE;
         }
      PQclear (tmp_res);
      }

   /*
    * Fill in the severity menu.
    */
   if (!in_error)
      {
      g_string_printf (sql_buffer, SELECT_SEVERITIES, pk_num);
      tmp_res = PQexec (conn, sql_buffer->str);
      sel = 0;
      targ_pk = atoi (PQgetvalue (main_res, 0, SEVERITY_NUM_POS));
      if (chk_sql_error (tmp_res, "Filling the severities list"))
         {
         menu = gtk_menu_new ();
         n = PQntuples (tmp_res);
         for (i = 0; i < n; i++)
            {
            pk = atoi (PQgetvalue (tmp_res, i, PK_NUM_POS));
            sel = (pk == targ_pk) ? i : sel;
            pk_list = g_list_append (pk_list, GINT_TO_POINTER (pk));
            item = gtk_menu_item_new_with_label
               (PQgetvalue (tmp_res, i, NAME_POS));
            gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
            gtk_widget_show (item);
            }
         g_object_set_data (G_OBJECT (severity_menu), PK_LIST, pk_list);
         pk_list = NULL;
         gtk_option_menu_set_menu (severity_menu, menu);
         gtk_option_menu_set_history (severity_menu, sel);
         }
      else
         {
         in_error = TRUE;
         }
      PQclear (tmp_res);
      }

   /*
    * Populate the responsible list, and attach the id list to the
    * option menu so we can later retrieve the ID of the person chosen.
    */
   if (!in_error)
      {
      g_string_printf (sql_buffer, SELECT_RESPONSIBLE, pk_num);
      tmp_res = PQexec (conn, sql_buffer->str);
      sel = 0;
      resp_id = g_string_new (PQgetvalue (main_res, 0, RESPONSIBLE_ID_POS));
      if (!chk_sql_error (tmp_res, "Getting responsible list"))
         {      
         in_error = TRUE;
         }
      else
         {
         menu = gtk_menu_new ();
         sel = 0;
         /* 
          * Get all of the names.  Create a list of login id's.  This
          * list MUST be in the same order as the associated option
          * menu, and will be destroyed when the dialog is destroyed.
          *
          * First load the NOBODY person.
          */
         id = g_string_new (RESP_NOBODY_ID);
         pk_list = g_list_append (pk_list, id);
         item = gtk_menu_item_new_with_label (RESP_NOBODY_NAME);
         gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
         gtk_widget_show (item);

         /* 
          * get all of the actual people
          */
         n = PQntuples (tmp_res);
         for (i = 0; i < n; i++)
            {
            id = g_string_new (PQgetvalue (tmp_res, i, RESPONSIBLE_ID_POS));
            pk_list = g_list_append (pk_list, id);
            sel = g_string_equal (id, resp_id) ? i + 1 : sel;
            item = gtk_menu_item_new_with_label
               (PQgetvalue (tmp_res, i, FULL_NAME_POS));
            gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
            gtk_widget_show (item);
            }
         g_object_set_data (G_OBJECT (responsible_menu), PK_LIST, pk_list);
         pk_list = NULL;
         gtk_option_menu_set_menu (responsible_menu, menu);
         gtk_option_menu_set_history (responsible_menu, sel);
         } /* end else */
      PQclear (tmp_res);
      g_string_free (resp_id, TRUE);
      }

   if (!in_error)
      {
      g_string_printf (sql_buffer, SELECT_VERSIONS, pk_num, pk_num);
      tmp_res = PQexec (conn, sql_buffer->str);
      if (!chk_sql_error (tmp_res, "Getting version list"))
         {      
         in_error = TRUE;
         }
      else
         {
         menu = gtk_menu_new ();

         /*
          * Start with the NULL version.
          */
         pk_list = g_list_append (pk_list, GINT_TO_POINTER (INVALID_PK));
         item = gtk_menu_item_new_with_label (NULL_NAME);
         gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
         gtk_widget_show (item);

         if (PQgetisnull (main_res, 0, FIX_VER_POS))
            {
            targ_pk = 0;
            }
         else
            {
            targ_pk = atoi (PQgetvalue (main_res, 0, FIX_VER_POS));
            }
         sel = 0;
         n = PQntuples (tmp_res);
         for (i = 0; i < n; i++)
            {
            pk = atoi (PQgetvalue (tmp_res, i, VERSION_NUM_POS));
            sel = (pk == targ_pk) ? i + 1 : sel;
            pk_list = g_list_append (pk_list, GINT_TO_POINTER (pk));
            item = gtk_menu_item_new_with_label
               (PQgetvalue (tmp_res, i, VERSION_TXT_POS));
            gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
            gtk_widget_show (item);
            }
         g_object_set_data (G_OBJECT (fix_ver_menu), PK_LIST, pk_list);
         pk_list = NULL;
         gtk_option_menu_set_menu (fix_ver_menu, menu);
         gtk_option_menu_set_history (fix_ver_menu, sel);
         }
      PQclear (tmp_res);
      }

   /*
    * TODO: Figure out what I wanted to do here, and do it.
    *
    * If the current user is in the submitter list, make the description
    * editable.  If the current user is in the reponsible list, make the
    * responsible list pickable and the fix description editable.
    */ 

   /*
    * If an error occurred along the way, destroy the dialog box.
    */
   if (in_error)
      {
      ROLLBACK (conn);
      gtk_widget_destroy (widget);
      }
   else
      {
      COMMIT (conn);
      set_all_clean (G_OBJECT (widget));
      gtk_dialog_set_response_sensitive (GTK_DIALOG (widget),
                                         GTK_RESPONSE_APPLY, FALSE);
      }
   if (main_res != NULL)
      {
      PQclear (main_res);
      }
   g_string_free (sql_buffer, TRUE);
}


static void 
on_pr_update_dlg_destroy (GObject *object, gpointer  user_data)
{
   GtkWidget     *main_widget;
   GtkOptionMenu *opt;
   GList         *temp_list;
   GList         *list_iter;
   PGresult      *res;
   GString       *sql_buffer;
   PGconn        *conn;
   gint           pk_num;

   main_widget = GTK_WIDGET (object);

   conn = g_object_get_data (G_OBJECT (main_widget), MY_DB_CONN);
   g_assert (conn != NULL);
   pk_num = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (main_widget),
                                                PK_NUM));

   sql_buffer = g_string_new ("");
   START_TRANSACTION (conn);
   g_string_printf (sql_buffer, REMOVE_LOCK, pk_num);
   res = PQexec (conn, sql_buffer->str);
   chk_sql_error (res, "Removing lock");
   COMMIT (conn);
   PQclear (res);
   g_string_free (sql_buffer, TRUE);

   /*
    * Free up the ID lists that are attached to the option menus.
    */
   opt = GTK_OPTION_MENU (lookup_widget (main_widget, RESPONSIBLE_MENU));
   g_assert (opt != NULL);
   temp_list = g_object_get_data (G_OBJECT (opt), PK_LIST);
   for (list_iter = g_list_first (temp_list);
         list_iter != NULL;
         list_iter = list_iter->next)
      {
      g_string_free ((GString *)list_iter->data, TRUE);
      }
   g_list_free (temp_list);

   opt = GTK_OPTION_MENU (lookup_widget (main_widget, SEVERITY_MENU));
   g_assert (opt != NULL);
   temp_list = g_object_get_data (G_OBJECT (opt), PK_LIST);
   g_list_free (temp_list);

   opt = GTK_OPTION_MENU (lookup_widget (main_widget, STATUS_MENU));
   g_assert (opt != NULL);
   temp_list = g_object_get_data (G_OBJECT (opt), PK_LIST);
   g_list_free (temp_list);

   opt = GTK_OPTION_MENU (lookup_widget (main_widget, FIX_VER_MENU));
   g_assert (opt != NULL);
   temp_list = g_object_get_data (G_OBJECT (opt), PK_LIST);
   g_list_free (temp_list);

   destroy_dirty_list (G_OBJECT (main_widget));
}


static gboolean 
apply_changes (GtkDialog *dlg)
{
   GtkWidget         *main_widget;
   GtkWidget         *parent;  /* main application window */
   GtkEntry          *subject_entry;
   GtkOptionMenu     *responsible_menu;
   GtkOptionMenu     *status_menu;
   GtkOptionMenu     *severity_menu;
   GtkOptionMenu     *fix_ver_menu;
   gint               sel;
   GtkTextView       *descr_text;
   GtkTextView       *fix_text;
   GtkTextView       *change_log_text;
   GtkTextBuffer     *buff;
   GtkTextIter        start;
   GtkTextIter        end;
   GList             *pk_list;
   GList             *list_iter;
   GString           *tmp_str;
   GString           *sql_buffer;
   PGresult          *res;
   gboolean           in_error = FALSE;
   gboolean           first = TRUE;
   gchar              num_buffer[NUM_BUFFER_LEN];
   gchar             *char_buf;
   PGconn            *conn;
   gint               pk_num;  /* the pk_num of the problem report */
   gint               pk;      /* temporary PK used while building update */
   GtkWidget         *msg_dlg;

   main_widget = GTK_WIDGET (dlg);
   conn = g_object_get_data (G_OBJECT (main_widget), MY_DB_CONN);
   g_assert (conn != NULL);
   pk_num = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (main_widget),
                                                PK_NUM));
   parent = lookup_widget (main_widget, MY_PARENT);

   /*
    * Get pointers to all the widgets we will need to access.
    */
   responsible_menu = 
      GTK_OPTION_MENU (lookup_widget (main_widget, RESPONSIBLE_MENU));
   subject_entry = 
      GTK_ENTRY (lookup_widget (main_widget, SUBJECT_ENTRY));
   status_menu = 
      GTK_OPTION_MENU (lookup_widget (main_widget, STATUS_MENU));
   severity_menu =
      GTK_OPTION_MENU (lookup_widget (main_widget, SEVERITY_MENU));
   descr_text = GTK_TEXT_VIEW (lookup_widget (main_widget, DESCR_EDIT));
   fix_text = GTK_TEXT_VIEW (lookup_widget (main_widget, FIX_DESCR_EDIT)); 
   change_log_text = GTK_TEXT_VIEW (lookup_widget (main_widget,
                                                   CHANGE_LOG_TEXT)); 
   fix_ver_menu = GTK_OPTION_MENU (lookup_widget (main_widget, FIX_VER_MENU));

   /*
    * Round up all of the information that we will need.
    *
    * Do validations as needed.
    */
   if (problem_update_dlg_dirty (main_widget))
      {
      sql_buffer = g_string_new ("UPDATE PROBLEM_REPORT SET ");
      if (!in_error && object_is_dirty (G_OBJECT (responsible_menu)))
         {
         if (!first)
            {
            sql_buffer = g_string_append (sql_buffer, ", ");
            }
         first = FALSE;
         sel = gtk_option_menu_get_history (responsible_menu);
         pk_list = g_object_get_data (G_OBJECT (responsible_menu), PK_LIST);
         list_iter = g_list_nth (pk_list, sel);
         tmp_str = g_string_new (((GString*)list_iter->data)->str);
         g_string_prepare_db_instr (tmp_str);
         sql_buffer = g_string_append (sql_buffer, "responsible_id = ");
         sql_buffer = g_string_append (sql_buffer, tmp_str->str);
         g_string_free (tmp_str, TRUE);
         }

      if (!in_error && object_is_dirty (G_OBJECT (subject_entry)))
         {
         tmp_str = g_string_new (gtk_entry_get_text (subject_entry));
         g_string_prepare_db_instr (tmp_str);
         /*
          * Subtract 2 to compensate for single quotes added for the
          * update statement.
          */
         if (tmp_str->len - 2 < MIN_SUBJECT_LEN ||
               (MIN_SUBJECT_LEN > 0 && !strcmp (tmp_str->str, "NULL")))
            {
            GString *msg_str;

            msg_str = g_string_new ("");
            g_string_printf
               (msg_str,
                _("You must enter at least %d character(s) in the subject"),
                MIN_SUBJECT_LEN);
            msg_dlg = gtk_message_dialog_new (GTK_WINDOW (dlg), 0,
                                              GTK_MESSAGE_ERROR,
                                              GTK_BUTTONS_OK,
                                              msg_str->str);
            gtk_dialog_run (GTK_DIALOG (msg_dlg));
            gtk_widget_destroy (msg_dlg);
            g_string_free (msg_str, TRUE);
            in_error = TRUE;
            }
         else
            {
            if (!first)
               {
               sql_buffer = g_string_append (sql_buffer, ", ");
               }
            first = FALSE;
            sql_buffer = g_string_append (sql_buffer, "title = ");
            sql_buffer = g_string_append (sql_buffer, tmp_str->str);
            }
         g_string_free (tmp_str, TRUE);
         }
   
      buff = gtk_text_view_get_buffer (descr_text);
      if (!in_error && gtk_text_buffer_get_modified (buff))
         {
         gtk_text_buffer_get_bounds (buff, &start, &end);
         char_buf = gtk_text_buffer_get_text (buff, &start, &end, FALSE);
         tmp_str = g_string_new (char_buf);
         g_free (char_buf);
         g_string_prepare_db_instr (tmp_str);
         /*
          * Subtract 2 to compensate for single quotes added for the
          * update statement.
          */
         if (tmp_str->len - 2 < MIN_DESCR_LEN ||
               (MIN_DESCR_LEN > 0 && !strcmp (tmp_str->str, "NULL")))
            {
            GString *msg_str;

            msg_str = g_string_new ("");
            g_string_printf
               (msg_str,
                _("You must enter at least %d character(s) in the description"),
                MIN_DESCR_LEN);
            msg_dlg = gtk_message_dialog_new (GTK_WINDOW (dlg), 0,
                                              GTK_MESSAGE_ERROR,
                                              GTK_BUTTONS_OK,
                                              msg_str->str);
            gtk_dialog_run (GTK_DIALOG (msg_dlg));
            gtk_widget_destroy (msg_dlg);
            g_string_free (msg_str, TRUE);
            in_error = TRUE;
            }
         else
            {
            if (!first)
               {
               sql_buffer = g_string_append (sql_buffer, ", ");
               }
            first = FALSE;
            sql_buffer = g_string_append (sql_buffer, "descr = ");
            sql_buffer = g_string_append (sql_buffer, tmp_str->str);
            }
         g_string_free (tmp_str, TRUE);
         }

      buff = gtk_text_view_get_buffer (fix_text);
      if (!in_error && gtk_text_buffer_get_modified (buff))
         {
         if (!first)
            {
            sql_buffer = g_string_append (sql_buffer, ", ");
            }
         first = FALSE;
         gtk_text_buffer_get_bounds (buff, &start, &end);
         char_buf = gtk_text_buffer_get_text (buff, &start, &end, FALSE);
         tmp_str = g_string_new (char_buf);
         g_free (char_buf);
         g_string_prepare_db_instr (tmp_str);
         sql_buffer = g_string_append (sql_buffer, "fix = ");
         sql_buffer = g_string_append (sql_buffer, tmp_str->str);
         g_string_free (tmp_str, TRUE);
         }

      buff = gtk_text_view_get_buffer (change_log_text);
      if (!in_error && gtk_text_buffer_get_modified (buff))
         {
         if (!first)
            {
            sql_buffer = g_string_append (sql_buffer, ", ");
            }
         first = FALSE;
         gtk_text_buffer_get_bounds (buff, &start, &end);
         char_buf = gtk_text_buffer_get_text (buff, &start, &end, FALSE);
         tmp_str = g_string_new (char_buf);
         g_free (char_buf);
         g_string_prepare_db_instr (tmp_str);
         sql_buffer = g_string_append (sql_buffer, "change_log_text = ");
         sql_buffer = g_string_append (sql_buffer, tmp_str->str);
         g_string_free (tmp_str, TRUE);
         }

      if (!in_error && object_is_dirty (G_OBJECT (fix_ver_menu)))
         {
         if (!first)
            {
            sql_buffer = g_string_append (sql_buffer, ", ");
            }
         first = FALSE;
         sel = gtk_option_menu_get_history (fix_ver_menu);
         pk_list = g_object_get_data (G_OBJECT (fix_ver_menu), PK_LIST);
         list_iter = g_list_nth (pk_list, sel);
         pk = GPOINTER_TO_INT (list_iter->data);
         sql_buffer = g_string_append (sql_buffer, "close_version = ");
         if (pk == INVALID_PK)
            {
            sql_buffer = g_string_append (sql_buffer, "NULL");
            }
         else
            {
            g_string_append_printf (sql_buffer, "%d", pk);
            }
         }

      if (!in_error && object_is_dirty (G_OBJECT (status_menu)))
         {
         if (!first)
            {
            sql_buffer = g_string_append (sql_buffer, ", ");
            }
         first = FALSE;
         sel = gtk_option_menu_get_history (status_menu);
         pk_list = g_object_get_data (G_OBJECT (status_menu), PK_LIST);
         list_iter = g_list_nth (pk_list, sel);
         pk = GPOINTER_TO_INT (list_iter->data);
         g_string_append_printf (sql_buffer, "status_num = %d", pk);
         }

      if (!in_error && object_is_dirty (G_OBJECT (severity_menu)))
         {
         if (!first)
            {
            sql_buffer = g_string_append (sql_buffer, ", ");
            }
         first = FALSE;
         sel = gtk_option_menu_get_history (severity_menu);
         pk_list = g_object_get_data (G_OBJECT (status_menu), PK_LIST);
         list_iter = g_list_nth (pk_list, sel);
         pk = GPOINTER_TO_INT (list_iter->data);
         g_string_append_printf (sql_buffer, "severity_num = %d", pk);
         }

      /* 
       * Update the database.
       */
      if (!in_error)
         {
         /*
          * If we are here, their ought to be at least one thing that
          * we are updating...
          */
         g_assert (!first);
         START_TRANSACTION (conn);
         sprintf (num_buffer, "%d", pk_num);
         sql_buffer = g_string_append (sql_buffer, " WHERE problem_num = ");
         sql_buffer = g_string_append (sql_buffer, num_buffer);
         res = PQexec (conn, sql_buffer->str);

         in_error = !chk_sql_error (res, "Updating Problem Report");
         PQclear (res);
         if (!in_error)
            {
            COMMIT (conn);
            set_all_clean (G_OBJECT (dlg));
            refresh_pr_list (parent);
            }
         else
            {
            ROLLBACK (conn);
            }
         }
      g_string_free (sql_buffer, TRUE);
      }

   return !in_error;
}


static void
on_ok_button_clicked (GtkDialog *dlg)
{
   GConfClient *client;

   client = g_object_get_data (G_OBJECT (dlg), MY_GCONF_CLIENT);
   g_assert (client != NULL);

   if (apply_changes (dlg))
      {
      save_dlg_size (client, GTK_WIDGET (dlg), PROB_UPD_SEC);
      gtk_widget_destroy (GTK_WIDGET (dlg));
      }
}


static void
on_apply_button_clicked (GtkDialog *dlg)
{
   apply_changes (dlg);
   gtk_dialog_set_response_sensitive (dlg, GTK_RESPONSE_APPLY, FALSE);
}


static void 
on_cancel_button_clicked (GtkDialog *dlg)
{
   GConfClient *client;

   client = g_object_get_data (G_OBJECT (dlg), MY_GCONF_CLIENT);
   g_assert (client != NULL);

   save_dlg_size (client, GTK_WIDGET (dlg), PROB_UPD_SEC);
   gtk_widget_destroy (GTK_WIDGET (dlg));
}

static void
on_pr_update_dlg_clicked (GtkDialog *dlg,
                          gint       btn_cd,
                          gpointer   user_data)
{
   switch (btn_cd)
      {
      case GTK_RESPONSE_OK:
         on_ok_button_clicked (dlg);
         break;
      case GTK_RESPONSE_APPLY:
         on_apply_button_clicked (dlg);
         break;
      case GTK_RESPONSE_CANCEL:
         on_cancel_button_clicked (dlg);
         break;
      case GTK_RESPONSE_DELETE_EVENT:
         break;
      default:
         g_assert_not_reached ();
      }
}


static void
check_chglog_tab_text (GtkTextBuffer *buff, gpointer user_data)
{
   GtkWidget     *main_win = GTK_WIDGET (user_data);
   GtkLabel      *tab_lbl;
   gint           curr_len;
   gint           stored_len;

   tab_lbl = GTK_LABEL (lookup_widget (main_win, CHGLOG_TAB_LBL));
   g_assert (tab_lbl != NULL);

   curr_len = gtk_text_buffer_get_char_count (buff);
   if (curr_len == 0)
      {
      gtk_label_set_text (tab_lbl, CHGLOG_DESCR);
      } 
   else
      {
      /*
       * Each time these things change, the current len is stored in
       * CHGLOG_TAB_CHAR_CNT.  We only want to change to a '*' if the 
       * last time we did not have any chars, and now we do...
       */
      stored_len = GPOINTER_TO_INT
         (g_object_get_data (G_OBJECT (main_win), CHGLOG_TAB_CHAR_CNT));
      if (!stored_len)
         {
         gtk_label_set_text (tab_lbl, CHGLOG_DESCR_STAR);
         }
      }

   g_object_set_data (G_OBJECT (main_win), CHGLOG_TAB_CHAR_CNT,
                      GINT_TO_POINTER (curr_len));
}


static void
check_fix_tab_text (GtkTextBuffer *buff, gpointer user_data)
{
   GtkWidget     *main_win = GTK_WIDGET (user_data);
   GtkLabel      *tab_lbl;
   gint           curr_len;
   gint           stored_len;

   tab_lbl = GTK_LABEL (lookup_widget (main_win, FIX_TAB_LBL));
   g_assert (tab_lbl != NULL);

   curr_len = gtk_text_buffer_get_char_count (buff);
   if (curr_len == 0)
      {
      gtk_label_set_text (tab_lbl, FIX_DESCR);
      } 
   else
      {
      /*
       * Each time these things change, the current len is stored in
       * FIX_TAB_CHAR_CNT.  We only want to change to a '*' if the 
       * last time we did not have any chars, and now we do...
       */
      stored_len = GPOINTER_TO_INT
         (g_object_get_data (G_OBJECT (main_win), FIX_TAB_CHAR_CNT));
      if (!stored_len)
         {
         gtk_label_set_text (tab_lbl, FIX_DESCR_STAR);
         }
      }

   g_object_set_data (G_OBJECT (main_win), FIX_TAB_CHAR_CNT,
                      GINT_TO_POINTER (curr_len));
}


static void
on_text_buffer_changed (GtkTextBuffer *buff, gpointer user_data)
{
   gtk_dialog_set_response_sensitive (GTK_DIALOG (user_data),
                                      GTK_RESPONSE_APPLY, TRUE);
}


static void
handle_widget_change (GtkWidget *widget, gpointer user_data)
{
   set_dirty (G_OBJECT (widget), NULL);
   gtk_dialog_set_response_sensitive (GTK_DIALOG (user_data),
                                      GTK_RESPONSE_APPLY, TRUE);
}


gboolean
problem_update_dlg_dirty (GtkWidget *dlg)
{
   GtkTextView   *view;
   GtkTextBuffer *buff;

   /*
    * First check the text view widgets, then check the list of other
    * widgets that can make the dialog dirty.
    */
   view = GTK_TEXT_VIEW (lookup_widget (dlg, DESCR_EDIT));
   g_assert (view != NULL);
   buff = gtk_text_view_get_buffer (view);
   if (gtk_text_buffer_get_modified (buff))
      {
      return TRUE;
      }

   view = GTK_TEXT_VIEW (lookup_widget (dlg, FIX_DESCR_EDIT));
   g_assert (view != NULL);
   buff = gtk_text_view_get_buffer (view);
   if (gtk_text_buffer_get_modified (buff))
      {
      return TRUE;
      }

   view = GTK_TEXT_VIEW (lookup_widget (dlg, CHANGE_LOG_TEXT));
   g_assert (view != NULL);
   buff = gtk_text_view_get_buffer (view);
   if (gtk_text_buffer_get_modified (buff))
      {
      return TRUE;
      }

   return objects_children_are_dirty (G_OBJECT (dlg));
}


GtkWidget *
create_problem_update_dlg (GtkWidget *parent,
                           PGconn    *conn,
                           gint       pr_pk)
{
   GtkWidget     *pr_update_dlg;
   GConfClient   *client;
   GString       *window_title;
   GString       *static_title;
   GtkWidget     *dialog_vbox;
   GtkWidget     *vbox;
   GtkWidget     *hbox;
   GtkWidget     *dtl_vbox;
   GtkWidget     *dtl_hbox;
   GtkWidget     *label;
   GtkWidget     *notebook;
   GtkWidget     *descr_scroll;
   GtkWidget     *descr_text;
   GtkWidget     *fix_scroll;
   GtkWidget     *fix_text;
   GtkWidget     *change_log_scroll;
   GtkWidget     *change_log_text;
   GtkWidget     *audit_trail_scroll;
   GtkWidget     *audit_trail_text;
   GtkWidget     *assign_menu;
   GtkWidget     *status_menu;
   GtkWidget     *severity_menu;
   GtkWidget     *version_menu;
   GtkWidget     *subject_entry;
   GtkSizeGroup  *size_group;
   GtkTextBuffer  *buff;

   client = g_object_get_data (G_OBJECT (parent), MY_GCONF_CLIENT);
   g_assert (client != NULL);

   window_title = g_string_new ("");
   g_string_printf (window_title, _("Edit Problem Report #%d"), pr_pk);
   pr_update_dlg = gtk_dialog_new_with_buttons
      (window_title->str,
       NULL, 0,
       GTK_STOCK_OK,
       GTK_RESPONSE_OK,
       GTK_STOCK_APPLY,
       GTK_RESPONSE_APPLY,
       GTK_STOCK_CANCEL,
       GTK_RESPONSE_CANCEL,
       NULL);
   g_string_free (window_title, TRUE);
   GLADE_HOOKUP_OBJECT_NO_REF (pr_update_dlg, pr_update_dlg, "pr_update_dlg");
   gtk_container_set_border_width (GTK_CONTAINER (pr_update_dlg),
                                   STD_DLG_BORDER);
   g_signal_connect (pr_update_dlg, "show",
                     G_CALLBACK (on_pr_update_dlg_show), NULL);
   g_signal_connect (pr_update_dlg, "destroy",
                     G_CALLBACK (on_pr_update_dlg_destroy), NULL);
   g_signal_connect (pr_update_dlg, "response",
                     G_CALLBACK (on_pr_update_dlg_clicked), NULL);
   gtk_window_set_resizable (GTK_WINDOW (pr_update_dlg), TRUE);
   gtk_dialog_set_default_response (GTK_DIALOG (pr_update_dlg),
                                    GTK_RESPONSE_OK);

   g_object_set_data (G_OBJECT (pr_update_dlg), TOP_DOG, pr_update_dlg);
   g_object_set_data (G_OBJECT (pr_update_dlg), MY_PARENT, parent);
   g_object_set_data (G_OBJECT (pr_update_dlg), MY_DB_CONN, conn);
   g_object_set_data (G_OBJECT (pr_update_dlg), MY_GCONF_CLIENT, client);
   g_object_set_data (G_OBJECT (pr_update_dlg), PK_NUM,
                      GINT_TO_POINTER (pr_pk));
   set_dlg_size (client, pr_update_dlg, PROB_UPD_SEC, 625, 500);

   /*----------------------------------------------------------------------
    * General layout:
    *   dialog_vbox
    *     vbox -- One for each basic section, provides separation
    *       hbox -- 2 slots, spaces in left, dtl vbox in right
    *         dtl_vbox -- contains a dtl_hbox for each item in section
    *           dtl_hbox
    *           dtl_hbox
    *           ...
    *     vbox
    *       hbox
    *         dtl_vbox
    *           dtl_hbox
    *           dtl_hbox
    *           ...
    *     ...
    *--------------------------------------------------------------------*/
   dialog_vbox = GTK_DIALOG (pr_update_dlg)->vbox;
   GLADE_HOOKUP_OBJECT (pr_update_dlg, dialog_vbox, "dialog_vbox");
   gtk_box_set_spacing (GTK_BOX (dialog_vbox), MAIN_VBOX_SPACING);

   size_group = gtk_size_group_new (GTK_SIZE_GROUP_HORIZONTAL);

   /*----------------------------------------------------------------------
    * Start the stuff that cannot be edited. 
    *--------------------------------------------------------------------*/
   vbox = gtk_vbox_new (FALSE, STD_BOX_SPACING);
   GLADE_HOOKUP_OBJECT (pr_update_dlg, vbox, "statictxtvbox");
   gtk_box_pack_start (GTK_BOX (dialog_vbox), vbox, FALSE, FALSE, 0);

   static_title = g_string_new ("<span weight=\"bold\">");
   g_string_append_printf (static_title, "Problem Report: %d</span>", pr_pk);
   label = gtk_label_new (NULL);
   gtk_label_set_markup (GTK_LABEL (label), static_title->str);
   gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.0);
   GLADE_HOOKUP_OBJECT (pr_update_dlg, label, "staticlbl");
   gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);
   g_string_free (static_title, FALSE);
   
   hbox = gtk_hbox_new (FALSE, STD_BOX_SPACING);
   GLADE_HOOKUP_OBJECT (pr_update_dlg, hbox, "statictxthbox");
   gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);

   label  = gtk_label_new ("    ");
   GLADE_HOOKUP_OBJECT (pr_update_dlg, label, "spacelbl1");
   gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);

   dtl_vbox = gtk_vbox_new (FALSE, STD_BOX_SPACING);
   GLADE_HOOKUP_OBJECT (pr_update_dlg, dtl_vbox, "dtlvbox1");
   gtk_box_pack_start (GTK_BOX (hbox), dtl_vbox, FALSE, FALSE, 0);

   dtl_hbox = gtk_hbox_new (FALSE, STD_BOX_SPACING);
   GLADE_HOOKUP_OBJECT (pr_update_dlg, dtl_hbox, "prjhbox");
   gtk_box_pack_start (GTK_BOX (dtl_vbox), dtl_hbox, FALSE, FALSE, 0);

   label = gtk_label_new (NULL);
   gtk_label_set_markup (GTK_LABEL (label),
                         _("<span weight=\"bold\">Project:</span>"));
   GLADE_HOOKUP_OBJECT (pr_update_dlg, label, "prjlbl");
   gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.0);
   gtk_size_group_add_widget (size_group, label);
   gtk_box_pack_start (GTK_BOX (dtl_hbox), label, FALSE, FALSE, 0);

   label = gtk_label_new (NULL);
   GLADE_HOOKUP_OBJECT (pr_update_dlg, label, PROJECT_LABEL);
   gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.0);
   gtk_box_pack_start (GTK_BOX (dtl_hbox), label, FALSE, FALSE, 0);

   dtl_hbox = gtk_hbox_new (FALSE, STD_BOX_SPACING);
   GLADE_HOOKUP_OBJECT (pr_update_dlg, dtl_hbox, "verhbox");
   gtk_box_pack_start (GTK_BOX (dtl_vbox), dtl_hbox, FALSE, FALSE, 0);

   label = gtk_label_new (NULL);
   gtk_label_set_markup (GTK_LABEL (label),
                         _("<span weight=\"bold\">Version:</span>"));
   GLADE_HOOKUP_OBJECT (pr_update_dlg, label, "verlbl");
   gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.0);
   gtk_size_group_add_widget (size_group, label);
   gtk_box_pack_start (GTK_BOX (dtl_hbox), label, FALSE, FALSE, 0);

   label = gtk_label_new (NULL);
   GLADE_HOOKUP_OBJECT (pr_update_dlg, label, VERSION_LABEL);
   gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.0);
   gtk_box_pack_start (GTK_BOX (dtl_hbox), label, FALSE, FALSE, 0);

   dtl_hbox = gtk_hbox_new (FALSE, STD_BOX_SPACING);
   GLADE_HOOKUP_OBJECT (pr_update_dlg, dtl_hbox, "pthbox");
   gtk_box_pack_start (GTK_BOX (dtl_vbox), dtl_hbox, FALSE, FALSE, 0);

   label = gtk_label_new (NULL);
   gtk_label_set_markup (GTK_LABEL (label),
                         _("<span weight=\"bold\">Problem Type:</span>"));
   GLADE_HOOKUP_OBJECT (pr_update_dlg, label, "ptlbl");
   gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.0);
   gtk_size_group_add_widget (size_group, label);
   gtk_box_pack_start (GTK_BOX (dtl_hbox), label, FALSE, FALSE, 0);

   label = gtk_label_new (NULL);
   GLADE_HOOKUP_OBJECT (pr_update_dlg, label, PRBTYPE_LABEL);
   gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.0);
   gtk_box_pack_start (GTK_BOX (dtl_hbox), label, FALSE, FALSE, 0);

   dtl_hbox = gtk_hbox_new (FALSE, STD_BOX_SPACING);
   GLADE_HOOKUP_OBJECT (pr_update_dlg, dtl_hbox, "subhbox");
   gtk_box_pack_start (GTK_BOX (dtl_vbox), dtl_hbox, FALSE, FALSE, 0);

   label = gtk_label_new (NULL);
   gtk_label_set_markup (GTK_LABEL (label),
                         _("<span weight=\"bold\">Submitter:</span>"));
   GLADE_HOOKUP_OBJECT (pr_update_dlg, label, "sublbl");
   gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.0);
   gtk_size_group_add_widget (size_group, label);
   gtk_box_pack_start (GTK_BOX (dtl_hbox), label, FALSE, FALSE, 0);

   label = gtk_label_new (NULL);
   GLADE_HOOKUP_OBJECT (pr_update_dlg, label, SUBMTTR_LABEL);
   gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.0);
   gtk_box_pack_start (GTK_BOX (dtl_hbox), label, FALSE, FALSE, 0);

   /*----------------------------------------------------------------------
    * Start the stuff that can be edited. 
    *--------------------------------------------------------------------*/
   vbox = gtk_vbox_new (FALSE, STD_BOX_SPACING);
   GLADE_HOOKUP_OBJECT (pr_update_dlg, vbox, "attrbvbox");
   gtk_box_pack_start (GTK_BOX (dialog_vbox), vbox, FALSE, FALSE, 0);

   hbox = gtk_hbox_new (FALSE, STD_BOX_SPACING);
   GLADE_HOOKUP_OBJECT (pr_update_dlg, hbox, "attrbhbox");
   gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);

   label  = gtk_label_new ("    ");
   GLADE_HOOKUP_OBJECT (pr_update_dlg, label, "spacelbl2");
   gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);

   dtl_vbox = gtk_vbox_new (FALSE, STD_BOX_SPACING);
   GLADE_HOOKUP_OBJECT (pr_update_dlg, dtl_vbox, "dtlvbox2");
   gtk_box_pack_start (GTK_BOX (hbox), dtl_vbox, TRUE, TRUE, 0);

   dtl_hbox = gtk_hbox_new (FALSE, STD_BOX_SPACING);
   GLADE_HOOKUP_OBJECT (pr_update_dlg, dtl_hbox, "asstohbox");
   gtk_box_pack_start (GTK_BOX (dtl_vbox), dtl_hbox, FALSE, FALSE, 0);

   label = gtk_label_new (_("Assigned To:"));
   GLADE_HOOKUP_OBJECT (pr_update_dlg, label, "asstolbl");
   gtk_size_group_add_widget (size_group, label);
   gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
   gtk_box_pack_start (GTK_BOX (dtl_hbox), label, FALSE, FALSE, 0);

   assign_menu = gtk_option_menu_new ();
   GLADE_HOOKUP_OBJECT (pr_update_dlg, assign_menu, RESPONSIBLE_MENU);
   gtk_box_pack_start (GTK_BOX (dtl_hbox), assign_menu, TRUE, TRUE, 0);
   g_signal_connect (assign_menu, "changed",
                     G_CALLBACK (handle_widget_change), pr_update_dlg);
   add_object_to_dirty_list (G_OBJECT (pr_update_dlg),
                             G_OBJECT (assign_menu));

   dtl_hbox = gtk_hbox_new (FALSE, STD_BOX_SPACING);
   GLADE_HOOKUP_OBJECT (pr_update_dlg, dtl_hbox, "stathbox");
   gtk_box_pack_start (GTK_BOX (dtl_vbox), dtl_hbox, FALSE, FALSE, 0);

   label = gtk_label_new (_("Status:"));
   GLADE_HOOKUP_OBJECT (pr_update_dlg, label, "statlbl");
   gtk_size_group_add_widget (size_group, label);
   gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
   gtk_box_pack_start (GTK_BOX (dtl_hbox), label, FALSE, FALSE, 0);

   status_menu = gtk_option_menu_new ();
   GLADE_HOOKUP_OBJECT (pr_update_dlg, status_menu, STATUS_MENU);
   gtk_box_pack_start (GTK_BOX (dtl_hbox), status_menu, TRUE, TRUE, 0);
   g_signal_connect (status_menu, "changed",
                     G_CALLBACK (handle_widget_change), pr_update_dlg);
   add_object_to_dirty_list (G_OBJECT (pr_update_dlg),
                             G_OBJECT (status_menu));

   dtl_hbox = gtk_hbox_new (FALSE, STD_BOX_SPACING);
   GLADE_HOOKUP_OBJECT (pr_update_dlg, dtl_hbox, "sevrhbox");
   gtk_box_pack_start (GTK_BOX (dtl_vbox), dtl_hbox, FALSE, FALSE, 0);

   label = gtk_label_new (_("Severity:"));
   GLADE_HOOKUP_OBJECT (pr_update_dlg, label, "sevrlbl");
   gtk_size_group_add_widget (size_group, label);
   gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
   gtk_box_pack_start (GTK_BOX (dtl_hbox), label, FALSE, FALSE, 0);

   severity_menu = gtk_option_menu_new ();
   GLADE_HOOKUP_OBJECT (pr_update_dlg, severity_menu, SEVERITY_MENU);
   gtk_box_pack_start (GTK_BOX (dtl_hbox), severity_menu, TRUE, TRUE, 0);
   g_signal_connect (severity_menu, "changed",
                     G_CALLBACK (handle_widget_change), pr_update_dlg);
   add_object_to_dirty_list (G_OBJECT (pr_update_dlg),
                             G_OBJECT (severity_menu));

   dtl_hbox = gtk_hbox_new (FALSE, STD_BOX_SPACING);
   GLADE_HOOKUP_OBJECT (pr_update_dlg, dtl_hbox, "targverhbox");
   gtk_box_pack_start (GTK_BOX (dtl_vbox), dtl_hbox, FALSE, FALSE, 0);

   label = gtk_label_new (_("Target Version:"));
   GLADE_HOOKUP_OBJECT (pr_update_dlg, label, "targverlbl");
   gtk_size_group_add_widget (size_group, label);
   gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
   gtk_box_pack_start (GTK_BOX (dtl_hbox), label, FALSE, FALSE, 0);

   version_menu = gtk_option_menu_new ();
   GLADE_HOOKUP_OBJECT (pr_update_dlg, version_menu, FIX_VER_MENU);
   gtk_box_pack_start (GTK_BOX (dtl_hbox), version_menu, TRUE, TRUE, 0);
   g_signal_connect (version_menu, "changed",
                     G_CALLBACK (handle_widget_change), pr_update_dlg);
   add_object_to_dirty_list (G_OBJECT (pr_update_dlg),
                             G_OBJECT (version_menu));

   dtl_hbox = gtk_hbox_new (FALSE, STD_BOX_SPACING);
   GLADE_HOOKUP_OBJECT (pr_update_dlg, dtl_hbox, "subjecthbox");
   gtk_box_pack_start (GTK_BOX (dtl_vbox), dtl_hbox, FALSE, FALSE, 0);

   label = gtk_label_new (_("Subject:"));
   GLADE_HOOKUP_OBJECT (pr_update_dlg, label, "subjectlbl");
   gtk_size_group_add_widget (size_group, label);
   gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
   gtk_box_pack_start (GTK_BOX (dtl_hbox), label, FALSE, FALSE, 0);

   subject_entry = gtk_entry_new ();
   GLADE_HOOKUP_OBJECT (pr_update_dlg, subject_entry, SUBJECT_ENTRY);
   gtk_box_pack_start (GTK_BOX (dtl_hbox), subject_entry, TRUE, TRUE, 0);
   g_signal_connect (subject_entry, "changed",
                     G_CALLBACK (handle_widget_change), pr_update_dlg);
   add_object_to_dirty_list (G_OBJECT (pr_update_dlg),
                             G_OBJECT (subject_entry));

   /*----------------------------------------------------------------------
    * Start the long descriptions (in tabbed windows).
    *--------------------------------------------------------------------*/
   notebook = gtk_notebook_new ();
   GLADE_HOOKUP_OBJECT (pr_update_dlg, notebook, "notebook");
   gtk_box_pack_start (GTK_BOX (dialog_vbox), notebook, TRUE, TRUE, 0);

   descr_scroll = gtk_scrolled_window_new (NULL, NULL);
   GLADE_HOOKUP_OBJECT (pr_update_dlg, descr_scroll, "descr_scroll");
   gtk_container_add (GTK_CONTAINER (notebook), descr_scroll);
   gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (descr_scroll), 
                                   GTK_POLICY_AUTOMATIC,
                                   GTK_POLICY_AUTOMATIC);
   gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (descr_scroll),
                                        GTK_SHADOW_ETCHED_IN);

   descr_text = gtk_text_view_new ();
   GLADE_HOOKUP_OBJECT (pr_update_dlg, descr_text, DESCR_EDIT);
   gtk_container_add (GTK_CONTAINER (descr_scroll), descr_text);
   gtk_widget_set_size_request (descr_text, -1, 150);
   gtk_text_view_set_editable (GTK_TEXT_VIEW (descr_text), TRUE);
   gtk_text_view_set_wrap_mode (GTK_TEXT_VIEW (descr_text), GTK_WRAP_WORD);
   buff = gtk_text_view_get_buffer (GTK_TEXT_VIEW (descr_text));
   g_signal_connect (buff, "changed", G_CALLBACK (on_text_buffer_changed),
                     pr_update_dlg);

   fix_scroll = gtk_scrolled_window_new (NULL, NULL);
   GLADE_HOOKUP_OBJECT (pr_update_dlg, fix_scroll, "fix_scroll");
   gtk_container_add (GTK_CONTAINER (notebook), fix_scroll);
   gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (fix_scroll), 
                                   GTK_POLICY_AUTOMATIC, 
                                   GTK_POLICY_AUTOMATIC);
   gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (fix_scroll),
                                        GTK_SHADOW_ETCHED_IN);

   fix_text = gtk_text_view_new ();
   GLADE_HOOKUP_OBJECT (pr_update_dlg, fix_text, FIX_DESCR_EDIT);
   gtk_container_add (GTK_CONTAINER (fix_scroll), fix_text);
   gtk_text_view_set_editable (GTK_TEXT_VIEW (fix_text), TRUE);
   gtk_text_view_set_wrap_mode (GTK_TEXT_VIEW (fix_text), GTK_WRAP_WORD);
   buff = gtk_text_view_get_buffer (GTK_TEXT_VIEW (fix_text));
   g_signal_connect (buff, "changed", G_CALLBACK (on_text_buffer_changed),
                     pr_update_dlg);
   g_signal_connect (buff, "changed", G_CALLBACK (check_fix_tab_text),
                     pr_update_dlg);

   change_log_scroll = gtk_scrolled_window_new (NULL, NULL);
   GLADE_HOOKUP_OBJECT (pr_update_dlg, change_log_scroll, "change_log_scroll");
   gtk_container_add (GTK_CONTAINER (notebook), change_log_scroll);
   gtk_widget_set_size_request (change_log_scroll, -1, 20);
   gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (change_log_scroll), 
                                   GTK_POLICY_AUTOMATIC, 
                                   GTK_POLICY_AUTOMATIC);
   gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (change_log_scroll),
                                        GTK_SHADOW_ETCHED_IN);

   change_log_text = gtk_text_view_new ();
   GLADE_HOOKUP_OBJECT (pr_update_dlg, change_log_text, CHANGE_LOG_TEXT);
   gtk_container_add (GTK_CONTAINER (change_log_scroll), change_log_text);
   gtk_text_view_set_editable (GTK_TEXT_VIEW (change_log_text), TRUE);
   gtk_text_view_set_wrap_mode (GTK_TEXT_VIEW (change_log_text), GTK_WRAP_WORD);
   buff = gtk_text_view_get_buffer (GTK_TEXT_VIEW (change_log_text));
   g_signal_connect (buff, "changed", G_CALLBACK (on_text_buffer_changed),
                     pr_update_dlg);
   g_signal_connect (buff, "changed", G_CALLBACK (check_chglog_tab_text),
                     pr_update_dlg);

   audit_trail_scroll = gtk_scrolled_window_new (NULL, NULL);
   GLADE_HOOKUP_OBJECT (pr_update_dlg, audit_trail_scroll,
                        "audit_trail_scroll");
   gtk_container_add (GTK_CONTAINER (notebook), audit_trail_scroll);
   gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (audit_trail_scroll), 
                                   GTK_POLICY_AUTOMATIC, 
                                   GTK_POLICY_AUTOMATIC);
   gtk_scrolled_window_set_shadow_type
      (GTK_SCROLLED_WINDOW (audit_trail_scroll), GTK_SHADOW_ETCHED_IN);

   audit_trail_text = gtk_text_view_new ();
   GLADE_HOOKUP_OBJECT (pr_update_dlg, audit_trail_text, AUDIT_TRAIL_EDIT);
   gtk_text_view_set_editable (GTK_TEXT_VIEW (audit_trail_text), FALSE);
   gtk_text_view_set_wrap_mode (GTK_TEXT_VIEW (audit_trail_text),
                                GTK_WRAP_WORD);
   gtk_container_add (GTK_CONTAINER (audit_trail_scroll), audit_trail_text);

   label = gtk_label_new (_("Description"));
   GLADE_HOOKUP_OBJECT (pr_update_dlg, label, "tab0lbl");
   gtk_notebook_set_tab_label 
      (GTK_NOTEBOOK (notebook),
       gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook), 0),
       label);
   gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);

   label = gtk_label_new (_("Fix Description"));
   GLADE_HOOKUP_OBJECT (pr_update_dlg, label, FIX_TAB_LBL);
   gtk_notebook_set_tab_label 
      (GTK_NOTEBOOK (notebook),
       gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook), 1),
       label);
   gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);

   label = gtk_label_new (_("ChangeLog Text"));
   GLADE_HOOKUP_OBJECT (pr_update_dlg, label, CHGLOG_TAB_LBL);
   gtk_notebook_set_tab_label 
      (GTK_NOTEBOOK (notebook),
       gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook), 2),
       label);
   gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);

   label = gtk_label_new (_("Audit Trail"));
   GLADE_HOOKUP_OBJECT (pr_update_dlg, label, "tab3lbl");
   gtk_notebook_set_tab_label 
      (GTK_NOTEBOOK (notebook),
       gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook), 3),
       label);
   gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);

   gtk_widget_show_all (pr_update_dlg);
   return pr_update_dlg;
}
